<?php

namespace Drupal\simple_sitemap_arbitrary_links\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Database\Connection;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\simple_sitemap\Manager\Generator as SitemapGenerator;
use Drupal\Core\Messenger\MessengerInterface;

/**
 * Provides a form for managing custom arbitrary sitemap links.
 */
class CustomSitemapLinksForm extends FormBase {
  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * The form builder.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected $formBuilder;

  /**
   * The simple sitemap generator.
   *
   * @var \Drupal\simple_sitemap\Manager\Generator
   */
  protected $sitemapGenerator;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'simple_sitemap_arbitrary_links_form';
  }

  /**
   * Constructs a new CustomSitemapLinksForm.
   *
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager.
   * @param \Drupal\Core\Form\FormBuilderInterface $form_builder
   *   The form builder.
   * @param \Drupal\simple_sitemap\Manager\Generator $sitemap_generator
   *   The simple sitemap generator.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   */
  public function __construct(Connection $database, RequestStack $request_stack, LanguageManagerInterface $language_manager, FormBuilderInterface $form_builder, SitemapGenerator $sitemap_generator, MessengerInterface $messenger) {
    $this->database = $database;
    $this->requestStack = $request_stack;
    $this->languageManager = $language_manager;
    $this->formBuilder = $form_builder;
    $this->sitemapGenerator = $sitemap_generator;
    $this->messenger = $messenger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
          $container->get('database'),
          $container->get('request_stack'),
          $container->get('language_manager'),
          $container->get('form_builder'),
          $container->get('simple_sitemap.generator'),
          $container->get('messenger')
      );
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $db = $this->database;
    $base_url = $this->requestStack->getCurrentRequest()->getSchemeAndHttpHost();

    // Fetch all existing custom sitemap links from the database, keyed by ID.
    try {
      $links = $db->select('custom_sitemap_links', 'c')
        ->fields('c', ['id', 'url', 'priority', 'changefreq', 'lastmod', 'langcode', 'external'])
        ->orderBy('id')
        ->execute()
        ->fetchAllAssoc('id');
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to load custom sitemap links: ' . $e->getMessage());
    }

    // Add a fieldset to group custom link configuration fields.
    $form['custom_links'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Custom Arbitrary Links'),
      '#tree' => TRUE,
      '#description' => $this->t('Define custom URLs for the XML sitemap, independent of Drupal entities.'),
      '#prefix' => '<div id="custom-links-wrapper">',
      '#suffix' => '</div>',
    ];

    $form['custom_links']['links'] = [
      '#type' => 'table',
      '#header' => [
        $this->t('URL'),
        $this->t('External'),
        $this->t('Priority'),
        $this->t('Change Frequency'),
        $this->t('Last Modified'),
        $this->t('Language'),
        $this->t('Operations'),
      ],
    ];

    // Retrieve available languages for the language dropdown.
    try {
      $languages = $this->languageManager->getLanguages();
      $lang_options = array_map(fn($lang) => $lang->getName(), $languages);
      $default_lang = $this->languageManager->getDefaultLanguage()->getId();
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to retrieve language options: ' . $e->getMessage());
    }

    // Populate the table with rows for each existing link.
    foreach ($links as $id => $link) {
      $link_url = $link->url;
      $form['custom_links']['links'][$id] = [
        '#prefix' => '<tr id="link-row-' . $id . '">',
        '#suffix' => '</tr>',
        'url' => [
          '#type' => 'textfield',
          '#default_value' => str_replace($base_url, '', $link_url),
          '#size' => 40,
        ],
        'external' => [
          '#type' => 'checkbox',
          '#default_value' => $link->external ?? 0,
        ],
        'priority' => [
          '#type' => 'select',
          '#default_value' => $link->priority ?? '1.0',
          '#options' => array_combine(
                    $p = ['0.1', '0.2', '0.3', '0.4', '0.5', '0.6', '0.7', '0.8', '0.9', '1.0'],
                    $p
          ),
        ],
        'changefreq' => [
          '#type' => 'select',
          '#default_value' => $link->changefreq ?? '',
          '#options' => ['' => $this->t('- Not specified -')] + array_map(
                        [$this, 't'],
                        ['always', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'never']
          ),
        ],
        'lastmod' => [
          '#type' => 'date',
          '#default_value' => $link->lastmod ? date('Y-m-d', strtotime($link->lastmod)) : '',
        ],
        'langcode' => [
          '#type' => 'select',
          '#default_value' => $link->langcode ?? $default_lang,
          '#options' => $lang_options,
        ],
        'remove' => [
          '#type' => 'submit',
          '#value' => $this->t('Remove'),
          '#name' => "remove_$id",
          '#delta' => $id,
          '#submit' => [[$this, 'removeLink']],
          '#attributes' => [
            'class' => ['remove-link'],
            'data-row-id' => "link-row-$id",
          ],
        ],
      ];
    }

    // Add a button to add a new link row.
    $form['custom_links']['add_link'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add new link'),
      '#ajax' => [
        'callback' => [$this, 'addLinkAjax'],
        'wrapper' => 'custom-links-wrapper',
        'event' => 'click',
      ],
    ];

    // Add the "Regenerate all sitemaps" checkbox.
    $form['simple_sitemap_regenerate_now'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Regenerate all sitemaps after saving'),
      '#default_value' => FALSE,
    ];

    // Add the submit button.
    $form['actions'] = [
      '#type' => 'actions',
      'submit' => [
        '#type' => 'submit',
        '#value' => $this->t('Save'),
      ],
    ];

    // Attach the JS library for client-side removal and scrolling.
    $form['#attached']['library'][] = 'simple_sitemap_arbitrary_links/remove_link';
    if ($form_state->get('scroll_to_custom_links')) {
      $form['#attached']['library'][] = 'simple_sitemap_arbitrary_links/scroll_to_custom_links';
    }

    return $form;
  }

  /**
   * AJAX callback to add a new link to the database and rebuild the form.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   The AJAX response object.
   */
  public function addLinkAjax(array &$form, FormStateInterface $form_state): AjaxResponse {
    $db = $this->database;
    $now = (new \DateTime('now', new \DateTimeZone('UTC')))->format('c');

    try {
      $new_id = $db->insert('custom_sitemap_links')
        ->fields([
          'url' => '',
          'priority' => '1.0',
          'changefreq' => '',
          'lastmod' => $now,
          'langcode' => $this->languageManager->getDefaultLanguage()->getId(),
          'external' => 0,
        ])
        ->execute();
      if (!$new_id) {
        throw new \RuntimeException('Failed to insert new sitemap link.');
      }
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Error adding new sitemap link: ' . $e->getMessage());
    }

    $form_state->setRebuild(TRUE);
    $rebuilt_form = $this->formBuilder->rebuildForm($this->getFormId(), $form_state, $form);
    return (new AjaxResponse())->addCommand(new ReplaceCommand('#custom-links-wrapper', $rebuilt_form['custom_links']));
  }

  /**
   * A handler to remove a link from the database and regenerate the sitemap.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function removeLink(array &$form, FormStateInterface $form_state): void {
    $trigger = $form_state->getTriggeringElement();
    $id = $trigger['#delta'] ?? NULL;

    if (!is_numeric($id)) {
      throw new \RuntimeException('Invalid link ID for removal.');
    }

    try {
      $deleted = $this->database->delete('custom_sitemap_links')
        ->condition('id', $id)
        ->execute();
      if ($deleted === 0) {
        throw new \RuntimeException("No link found with ID $id to remove.");
      }
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to remove sitemap link: ' . $e->getMessage());
    }

    // Regenerate the sitemap after removal.
    try {
      $this->sitemapGenerator->generate();
      $this->messenger->addStatus($this->t('Sitemap has been regenerated after removing link ID @id.', ['@id' => $id]));
    }
    catch (\Exception $e) {
      $this->messenger->addError($this->t('Failed to regenerate sitemap after removing link ID @id: @message', [
        '@id' => $id,
        '@message' => $e->getMessage(),
      ]));
    }

    $form_state->setValues([]);
    $form_state->setUserInput([]);
    $form_state->set('scroll_to_custom_links', TRUE);
    $form_state->setRebuild(TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $db = $this->database;
    $base_url = $this->requestStack->getCurrentRequest()->getSchemeAndHttpHost();
    $values = $form_state->getValue(['custom_links', 'links']) ?: [];

    foreach ($values as $id => $link) {
      if (!empty($link['url'])) {
        $url = $link['url'];

        if (empty($link['external'])) {
          $url = (strpos($url, '/') === 0 || preg_match('#^https?://#', $url)) ? $url : "/$url";
          $url = (strpos($url, '/') === 0) ? $base_url . $url : $url;
        }

        try {
          $lastmod = !empty($link['lastmod']) ? (new \DateTime($link['lastmod'], new \DateTimeZone('UTC')))->format('c') : NULL;
        }
        catch (\Exception $e) {
          throw new \RuntimeException('Invalid last modified date for link ID ' . $id . ': ' . $e->getMessage());
        }

        try {
          $db->merge('custom_sitemap_links')
            ->key('id', (string) $id)
            ->fields([
              'url' => $url,
              'priority' => $link['priority'],
              'changefreq' => $link['changefreq'],
              'lastmod' => $lastmod,
              'langcode' => $link['langcode'],
              'external' => $link['external'] ?? 0,
            ])
            ->execute();
        }
        catch (\Exception $e) {
          throw new \RuntimeException('Failed to save sitemap link ID ' . $id . ': ' . $e->getMessage());
        }
      }
    }

    // Regenerate the sitemap if the checkbox is checked.
    if ($form_state->getValue('simple_sitemap_regenerate_now')) {
      try {
        $this->sitemapGenerator->generate();
        $this->messenger->addStatus($this->t('Sitemap has been regenerated.'));
      }
      catch (\Exception $e) {
        $this->messenger->addError($this->t('Failed to regenerate sitemap: @message', ['@message' => $e->getMessage()]));
      }
    }

    $this->messenger->addStatus($this->t('Custom sitemap links have been saved.'));
  }

}
